/* Copyright 1986 Norm Hutchinson.	 May not be used for any  	   *
 * purpose without written permission from the author.			   */

#include "Kernel/h/assert.h"
#include "Kernel/h/system.h"
#include "Kernel/h/map.h"

#define DEBUGMAP
#define INITIALSIZE 16

#define HASH(key, map) ((((key)<< 1) ^ ((key) >> 6)) & map->maxIndex)
static void CheckOutHashTable();

#define MAXMAPSTANDBY 20
int              nextFreeMap = 0;
Map              standbyMaps[MAXMAPSTANDBY];

#ifdef DEBUGMAP
int inhibitCheck = 0;
#endif

Map Map_Create()
{
#ifdef xkernel
  int x = spl7();
#endif xkernel
  register int i;
  register Map m;
  register TEPtr    entryPtr;
  register int myNil = NIL;
  if (nextFreeMap > 0) {
    /* Grab one from standby stack */
    m           = standbyMaps[--nextFreeMap];
#ifdef xkernel
    splx(x);
#endif xkernel
    return m;
  }

  /* Allocate a new one */
  m = (Map) malloc(sizeof(MapRecord));
  m->table = (TEPtr) malloc(INITIALSIZE * sizeof(TE));
  m->maxIndex = INITIALSIZE - 1;
  m->maxCount = m->maxIndex - (m->maxIndex >> 2); /* Max fill 87.5 % */
  m->count = 0;
  i = m->maxIndex;
  entryPtr = &m->table[i+1];
  do {
      (--entryPtr)->key = myNil;
  } while (--i >= 0);
#ifdef DEBUGMAP
      CheckOutHashTable(m);
#endif DEBUGMAP
#ifdef xkernel
  splx(x);
#endif xkernel
  return(m);
}

Map Map_CreateSized(fCount)
int fCount;
{
#ifdef xkernel
  int x = spl7();
#endif xkernel
  register int i;
  register Map m;
  register TEPtr    entryPtr;
  register int myNil = NIL;
  
  assert(fCount > 0);
  m = (Map) malloc(sizeof(MapRecord));
  /* Make i the first power of two greater than fCount */
  for (i = 1; i <= fCount; i += i);
  i--;
  m->maxIndex = i;
  m->maxCount = m->maxIndex - (m->maxIndex >> 2); /* Max fill 87.5 % */
  m->count = 0;
  m->table = (TEPtr) malloc((unsigned)((m->maxIndex+1) * sizeof(TE)));
  entryPtr = &m->table[i+1];
  do {
      (--entryPtr)->key = myNil;
  } while (--i >= 0);
# ifdef lint
  CheckOutHashTable(m);
#endif  
#ifdef DEBUGMAP
  CheckOutHashTable(m);
#endif DEBUGMAP
#ifdef xkernel
  splx(x);
#endif xkernel
  return(m);
}

void Map_Clear(m)
register Map m;
{
  register int i;
  register TEPtr entryPtr;
  register int myNil = NIL;
  
  m->count      = 0;
  i             = m->maxIndex;
  entryPtr      = &m->table[i+1];
  do {
      (--entryPtr)->key = myNil;
  } while (--i >= 0);
}

void Map_Destroy(m)
register Map m;
{
  register int i;
  register TEPtr entryPtr;
  register int myNil = NIL;
#ifdef xkernel
  int x;
#endif xkernel
  
  if (m == (Map) NULL) return;
#ifdef xkernel
  x = spl7();
#endif xkernel
  if ((nextFreeMap < MAXMAPSTANDBY)
    && (m->maxIndex == INITIALSIZE - 1)
  ) {
      if (m->count != 0) {
	  /* Reinitialize it */
          m->count = 0;
          i = m->maxIndex;
	  entryPtr = &m->table[i+1];
	  do {
	      (--entryPtr)->key = myNil;
	  } while (--i >= 0);
      }
      
      /* Nice map, recycle it */
      standbyMaps[nextFreeMap++] = m;
#ifdef DEBUGMAP
      CheckOutHashTable(m);
#endif DEBUGMAP
#ifdef xkernel
      splx(x);
#endif xkernel
      return;
  }
#ifdef DEBUGMAP
  CheckOutHashTable(m);
#endif DEBUGMAP
  free((char *)m->table);
  free((char *)m);
#ifdef xkernel
  splx(x);
#endif xkernel
}

static void ExpandHashTable(m)
register Map m;
{
  register TE *nh, *oe, *ne;
  register int oldHashTableSize = m->maxIndex + 1, i;
  register int key;
  int index;

  m->maxIndex = oldHashTableSize * 2 - 1;
  m->maxCount = m->maxIndex - (m->maxIndex >> 2);
  nh = (TEPtr) malloc((unsigned)(oldHashTableSize * 2 * sizeof(TE)));
  for (i = 0; i <= m->maxIndex; i++) nh[i].key = NIL;
  for (i = 0; i < oldHashTableSize; i++) {
    oe = &m->table[i];
    key = oe->key;
    if (key == NIL) continue;
    index = HASH(key, m);
    while (1) {
      ne = &nh[index];
      if (ne->key == NIL) {
	ne->key = oe->key;
	ne->value = oe->value;
	break;
      } else {
	assert(ne->key !=key);
	index++;
	if (index > m->maxIndex) index = 0;
      }
    }
  }
  free((char *)m->table);
  m->table = nh;
#ifdef DEBUGMAP
  CheckOutHashTable(m);
#endif DEBUGMAP
}

int Map_Lookup(m, key)
register Map m;
register int key;
{
  register int index = HASH(key, m);
  register TEPtr e;
  register int myNil = NIL;
  register int limit = m->maxIndex;
  
#ifdef DEBUGMAP
  CheckOutHashTable(m);
#endif DEBUGMAP
  e = &m->table[index];
  while (1) {
    if (e->key == myNil) {		/* we did not find it */
      return (myNil);
    } else if (e->key == key) {
      return (e->value);
    }
    e++;
    if (++index > limit) {
	index = 0;
        e = &m->table[0];
    }
  }
}

int Map_InverseLookup(m, value)
register Map m;
register int value;
/* do the inverse mapping; if ambiguous return any one of possible values */
/* WARNING: Highly inefficient -- linear search */
{
  register TEPtr e, last;

#ifdef DEBUGMAP
  CheckOutHashTable(m);
#endif DEBUGMAP
  for (last = &m->table[m->maxIndex+1], e = &m->table[0]; e != last; e++) {
      if ((e->key != NIL) && (e->value == value)) return (e->key);
  }
  return ((int) NIL);
}

void Map_Delete(m, key)
register Map m;
register int key;
/* Remove the entry, if it is there */
{
#ifdef xkernel
  int x = spl7();
#endif xkernel
  register int index = HASH(key, m);
  register int value;
  register TEPtr e;

  while (1) {
    e = &m->table[index];
    if (e->key == NIL) {		/* we did not find it */
#ifdef DEBUGMAP
      CheckOutHashTable(m);
#endif DEBUGMAP
#ifdef xkernel
      splx(x);
#endif xkernel
      return;
    }
    if (e->key == key) {
      /* Found it, now remove it */
      m->count--;
      e->key = NIL;
      e->value = NIL;
#ifdef DEBUGMAP
      inhibitCheck = 1;
#endif
      while (1) {
	/* rehash until we reach nil again */
	if (++index > m->maxIndex) index = 0;
	e = & m->table[index];
	key = e->key;
	if (key == NIL) {
#ifdef DEBUGMAP
	    inhibitCheck = 0;
	    CheckOutHashTable(m);
#endif DEBUGMAP
#ifdef xkernel
	    splx(x);
#endif xkernel
	    return;
	}
	/* rehashing is done by removing then reinserting */
	value = e->value;
	e->key = e->value = NIL;
	m->count--;
	Map_Insert(m, key, value);
      }
    }
    if (++index > m->maxIndex) index = 0;
  }
}

void Map_Insert(m, key, value)
register Map m;
register int key, value;
{
  register int index;
  register TEPtr e;
#ifdef xkernel
  int x = spl7();
#endif xkernel
  assert(key != NIL);
#ifdef DEBUGMAP
  CheckOutHashTable(m);
#endif
  if (m->count >= m->maxCount) ExpandHashTable(m);
  index = HASH(key, m);
  while (1) {
    e = &m->table[index];
    if (e->key == NIL) {		/* put it here */
      e->key = key;
      e->value = value;
      m->count++;
#ifdef DEBUGMAP
      CheckOutHashTable(m);
#endif DEBUGMAP
#ifdef xkernel
      splx(x);
#endif xkernel
      return;
    } else if (e->key == key) {
      e->value = value;
#ifdef DEBUGMAP
      CheckOutHashTable(m);
#endif DEBUGMAP
#ifdef xkernel
      splx(x);
#endif xkernel
      return;
    }
    if (++index > m->maxIndex) index = 0;
  }
}

int Map_FindNext(m, startIndex, keyPtr, valuePtr)
register Map m;
int *startIndex;
int *keyPtr, *valuePtr;
{
  register int i;
  register TEPtr e;
#ifdef xkernel
  int x = spl7();
#endif xkernel

#ifdef DEBUGMAP
  CheckOutHashTable(m);
#endif
  for (i = *startIndex; i <= m->maxIndex; i++) {
    e = &m->table[i];
    if (e->key != NIL) {
      *keyPtr = e->key;
      *valuePtr = e->value;
      *startIndex = i + 1;
#ifdef xkernel
      splx(x);
#endif xkernel
      return(1);
    }
  }
  *keyPtr = NIL;
  *valuePtr = NIL;
  *startIndex = -1;
#ifdef xkernel
  splx(x);
#endif xkernel
  return(0);
}

void Map_Print(m)
register Map m;
{
    int key, value, index;
    printf(
	"\nDump of map @ 0x%05x, %d entr%s, max %d, maxIndex %d\nIndex\tKey\t\tValue\n",
	m, m->count, m->count == 1 ? "y" : "ies",  m->maxCount, m->maxIndex);
    for (index = 0; index <= m->maxIndex; index++) {
	key = m->table[index].key;
	value = m->table[index].value;
	printf("%3d\t0x%08.8x\t0x%08.8x\n", index, key, value);
    }
}

#define isReasonable(x) ((int)(x) > 0 && (int)(x) < 0x1000000)

static void CheckOutHashTable(m)
register Map m;
{
  register int i;
  register TEPtr realElement, e;
  register int index, firstIndex, count;
  count = 0;

#ifdef DEBUGMAP
  if (inhibitCheck) return;
#endif
  if (!isReasonable(m)) {
    printf("Invalid map pointer 0x%x\n", m);
    assert(0);
  }
  if (!isReasonable(m->table)) {
    printf("Invalid map table pointer 0x%x\n", m->table);
    assert(0);
  }
  for (i = 1; i < m->maxIndex; i <<= 1) ;
  if (i != m->maxIndex + 1) {
    printf("Weird maxindex = %d\n", m->maxIndex);
    assert(0);
  }
  for (i = 0; i <= m->maxIndex; i++) {
    realElement = &m->table[i];
    if (realElement->key != NIL) {
      count++;
      index = HASH(realElement->key, m);
      firstIndex = index;
      while (1) {
	e = &m->table[index];
	if (e->key == NIL) {		/* we did not find it */
	  break;
	} else if (e->key == realElement->key) {
	  break;
	} else {
	  index++;
	  if (index > m->maxIndex) index = 0;
	  if (index == firstIndex) {
	    index = -1;
	    break;
	  }
	}
      }
      
      if (index == -1 || e->key != realElement->key) {
	  printf(
	    "Map problem: Key 0x%x, rightIndex %d, realIndex %d value 0x%x\n",
	    realElement->key, firstIndex, i, realElement->value);
	  Map_Print(m);
	  assert(0);
      }
    }
  }  
  if (count != m->count) {
    printf("Map problem: Should have %d entries, but found %d.\n",
	   m->count, count);
    Map_Print(m);
    assert(0);
  }
}

